@@ -28,7 +28,6 @@ import types
 import warnings
 import zlib
 from collections import defaultdict
-from inspect import getfullargspec
 from io import StringIO
 from random import Random
 from typing import Any, BinaryIO, Callable, Hashable, List, Optional, TypeVar, Union
@@ -60,13 +59,12 @@ from hypothesis.errors import (
     Unsatisfiable,
     UnsatisfiedAssumption,
 )
-from hypothesis.executors import new_style_executor
+from hypothesis.executors import default_new_style_executor, new_style_executor
 from hypothesis.internal.compat import (
     PYPY,
     bad_django_TestCase,
     get_type_hints,
     int_from_bytes,
-    qualname,
 )
 from hypothesis.internal.conjecture.data import ConjectureData, Status
 from hypothesis.internal.conjecture.engine import ConjectureRunner
@@ -84,6 +82,7 @@ from hypothesis.internal.reflection import (
     define_function_signature,
     function_digest,
     get_pretty_function_description,
+    getfullargspec_except_self as getfullargspec,
     impersonate,
     is_mock,
     proxies,
@@ -226,7 +225,7 @@ class WithRunner(MappedSearchStrategy):
         return f"WithRunner({self.mapped_strategy!r}, runner={self.runner!r})"
 
 
-def is_invalid_test(name, original_argspec, given_arguments, given_kwargs):
+def is_invalid_test(test, original_argspec, given_arguments, given_kwargs):
     """Check the arguments to ``@given`` for basic usage constraints.
 
     Most errors are not raised immediately; instead we return a dummy test
@@ -240,6 +239,11 @@ def is_invalid_test(name, original_argspec, given_arguments, given_kwargs):
             raise InvalidArgument(message)
 
         wrapped_test.is_hypothesis_test = True
+        wrapped_test.hypothesis = HypothesisHandle(
+            inner_test=test,
+            get_fuzz_target=wrapped_test,
+            given_kwargs=given_kwargs,
+        )
         return wrapped_test
 
     if not (given_arguments or given_kwargs):
@@ -256,9 +260,9 @@ def is_invalid_test(name, original_argspec, given_arguments, given_kwargs):
     if len(given_arguments) > len(original_argspec.args):
         args = tuple(given_arguments)
         return invalid(
-            "Too many positional arguments for %s() were passed to @given "
-            "- expected at most %d arguments, but got %d %r"
-            % (name, len(original_argspec.args), len(args), args)
+            f"Too many positional arguments for {test.__name__}() were passed to "
+            f"@given - expected at most {int(len(original_argspec.args))} "
+            f"arguments, but got {int(len(args))} {args!r}"
         )
 
     if infer in given_arguments:
@@ -277,8 +281,8 @@ def is_invalid_test(name, original_argspec, given_arguments, given_kwargs):
     if extra_kwargs and not original_argspec.varkw:
         arg = extra_kwargs[0]
         return invalid(
-            "%s() got an unexpected keyword argument %r, from `%s=%r` in @given"
-            % (name, arg, arg, given_kwargs[arg])
+            f"{test.__name__}() got an unexpected keyword argument {arg!r}, "
+            f"from `{arg}={given_kwargs[arg]!r}` in @given"
         )
     if original_argspec.defaults or original_argspec.kwonlydefaults:
         return invalid("Cannot apply @given to a function with defaults.")
@@ -325,9 +329,8 @@ def execute_explicit_examples(state, wrapped_test, arguments, kwargs):
         if example.args:
             if len(example.args) > len(original_argspec.args):
                 raise InvalidArgument(
-                    "example has too many arguments for test. "
-                    "Expected at most %d but got %d"
-                    % (len(original_argspec.args), len(example.args))
+                    "example has too many arguments for test. Expected at most "
+                    f"{len(original_argspec.args)} but got {len(example.args)}"
                 )
             example_kwargs.update(
                 dict(zip(original_argspec.args[-len(example.args) :], example.args))
@@ -525,6 +528,7 @@ class StateForActualGivenExecution:
 
         self.files_to_propagate = set()
         self.failed_normally = False
+        self.failed_due_to_deadline = False
 
         self.explain_traces = defaultdict(set)
 
@@ -664,6 +668,8 @@ class StateForActualGivenExecution:
             trace = frozenset()
             if (
                 self.failed_normally
+                and not self.failed_due_to_deadline
+                and Phase.shrink in self.settings.phases
                 and Phase.explain in self.settings.phases
                 and sys.gettrace() is None
                 and not PYPY
@@ -732,6 +738,9 @@ class StateForActualGivenExecution:
                 if trace:  # pragma: no cover
                     # Trace collection is explicitly disabled under coverage.
                     self.explain_traces[interesting_origin].add(trace)
+                if interesting_origin[0] == DeadlineExceeded:
+                    self.failed_due_to_deadline = True
+                    self.explain_traces.clear()
                 data.mark_interesting(interesting_origin)
 
     def run_engine(self):
@@ -769,10 +778,8 @@ class StateForActualGivenExecution:
             )
         else:
             if runner.valid_examples == 0:
-                raise Unsatisfiable(
-                    "Unable to satisfy assumptions of hypothesis %s."
-                    % (get_pretty_function_description(self.test),)
-                )
+                rep = get_pretty_function_description(self.test)
+                raise Unsatisfiable(f"Unable to satisfy assumptions of {rep}")
 
         if not self.falsifying_examples:
             return
@@ -959,7 +966,7 @@ def given(
         original_argspec = getfullargspec(test)
 
         check_invalid = is_invalid_test(
-            test.__name__, original_argspec, given_arguments, given_kwargs
+            test, original_argspec, given_arguments, given_kwargs
         )
 
         # If the argument check found problems, return a dummy test function
@@ -971,10 +978,8 @@ def given(
         # positional arguments into keyword arguments for simplicity.
         if given_arguments:
             assert not given_kwargs
-            for name, strategy in zip(
-                reversed(original_argspec.args), reversed(given_arguments)
-            ):
-                given_kwargs[name] = strategy
+            nargs = len(given_arguments)
+            given_kwargs.update(zip(original_argspec.args[-nargs:], given_arguments))
         # These have been converted, so delete them to prevent accidental use.
         del given_arguments
 
@@ -993,8 +998,8 @@ def given(
                 def wrapped_test(*arguments, **kwargs):
                     __tracebackhide__ = True
                     raise InvalidArgument(
-                        "passed %s=infer for %s, but %s has no type annotation"
-                        % (name, test.__name__, name)
+                        f"passed {name}=infer for {test.__name__}, but {name} has "
+                        "no type annotation"
                     )
 
                 return wrapped_test
@@ -1011,15 +1016,12 @@ def given(
 
             if getattr(test, "is_hypothesis_test", False):
                 raise InvalidArgument(
-                    (
-                        "You have applied @given to the test %s more than once, which "
-                        "wraps the test several times and is extremely slow. A "
-                        "similar effect can be gained by combining the arguments "
-                        "of the two calls to given. For example, instead of "
-                        "@given(booleans()) @given(integers()), you could write "
-                        "@given(booleans(), integers())"
-                    )
-                    % (test.__name__,)
+                    f"You have applied @given to the test {test.__name__} more than "
+                    "once, which wraps the test several times and is extremely slow. "
+                    "A similar effect can be gained by combining the arguments "
+                    "of the two calls to given. For example, instead of "
+                    "@given(booleans()) @given(integers()), you could write "
+                    "@given(booleans(), integers())"
                 )
 
             settings = wrapped_test._hypothesis_internal_use_settings
@@ -1031,22 +1033,39 @@ def given(
             )
             arguments, kwargs, test_runner, search_strategy = processed_args
 
+            if (
+                inspect.iscoroutinefunction(test)
+                and test_runner is default_new_style_executor
+            ):
+                # See https://github.com/HypothesisWorks/hypothesis/issues/3054
+                # If our custom executor doesn't handle coroutines, or we return an
+                # awaitable from a non-async-def function, we just rely on the
+                # return_value health check.  This catches most user errors though.
+                raise InvalidArgument(
+                    "Hypothesis doesn't know how to run async test functions like "
+                    f"{test.__name__}.  You'll need to write a custom executor, "
+                    "or use a library like pytest-asyncio or pytest-trio which can "
+                    "handle the translation for you.\n    See https://hypothesis."
+                    "readthedocs.io/en/latest/details.html#custom-function-execution"
+                )
+
             runner = getattr(search_strategy, "runner", None)
             if isinstance(runner, TestCase) and test.__name__ in dir(TestCase):
                 msg = (
-                    "You have applied @given to the method %s, which is "
+                    f"You have applied @given to the method {test.__name__}, which is "
                     "used by the unittest runner but is not itself a test."
-                    "  This is not useful in any way." % test.__name__
+                    "  This is not useful in any way."
                 )
                 fail_health_check(settings, msg, HealthCheck.not_a_test_method)
             if bad_django_TestCase(runner):  # pragma: no cover
                 # Covered by the Django tests, but not the pytest coverage task
                 raise InvalidArgument(
-                    "You have applied @given to a method on %s, but this "
+                    "You have applied @given to a method on "
+                    f"{type(runner).__qualname__}, but this "
                     "class does not inherit from the supported versions in "
                     "`hypothesis.extra.django`.  Use the Hypothesis variants "
                     "to ensure that each example is run in a separate "
-                    "database transaction." % qualname(type(runner))
+                    "database transaction."
                 )
 
             state = StateForActualGivenExecution(
@@ -1275,8 +1294,8 @@ def find(
 
     if not isinstance(specifier, SearchStrategy):
         raise InvalidArgument(
-            "Expected SearchStrategy but got %r of type %s"
-            % (specifier, type(specifier).__name__)
+            f"Expected SearchStrategy but got {specifier!r} of "
+            f"type {type(specifier).__name__}"
         )
     specifier.validate()
 
