@@ -1,7 +1,7 @@
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis/
 #
-# Most of this work is copyright (C) 2013-2020 David R. MacIver
+# Most of this work is copyright (C) 2013-2021 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -13,7 +13,11 @@
 #
 # END HEADER
 
-from collections import Counter, defaultdict
+import math
+import sys
+import time
+from collections import defaultdict
+from contextlib import contextmanager
 from enum import Enum
 from random import Random, getrandbits
 from weakref import WeakKeyDictionary
@@ -22,6 +26,7 @@ import attr
 
 from hypothesis import HealthCheck, Phase, Verbosity, settings as Settings
 from hypothesis._settings import local_settings
+from hypothesis.errors import StopTest
 from hypothesis.internal.cache import LRUReusedCache
 from hypothesis.internal.compat import ceil, int_from_bytes
 from hypothesis.internal.conjecture.data import (
@@ -30,23 +35,17 @@ from hypothesis.internal.conjecture.data import (
     DataObserver,
     Overrun,
     Status,
-    StopTest,
 )
 from hypothesis.internal.conjecture.datatree import (
     DataTree,
     PreviouslyUnseenBehaviour,
     TreeRecordingObserver,
 )
-from hypothesis.internal.conjecture.junkdrawer import clamp
+from hypothesis.internal.conjecture.junkdrawer import clamp, stack_depth_of_caller
 from hypothesis.internal.conjecture.pareto import NO_SCORE, ParetoFront, ParetoOptimiser
 from hypothesis.internal.conjecture.shrinker import Shrinker, sort_key
 from hypothesis.internal.healthcheck import fail_health_check
-from hypothesis.reporting import base_report
-
-# Tell pytest to omit the body of this module from tracebacks
-# https://docs.pytest.org/en/latest/example/simple.html#writing-well-integrated-assertion-helpers
-__tracebackhide__ = True
-
+from hypothesis.reporting import base_report, report
 
 MAX_SHRINKS = 500
 CACHE_SIZE = 10000
@@ -64,11 +63,18 @@ class HealthCheckState:
 
 
 class ExitReason(Enum):
-    max_examples = 0
-    max_iterations = 1
-    max_shrinks = 3
-    finished = 4
-    flaky = 5
+    max_examples = "settings.max_examples={s.max_examples}"
+    max_iterations = (
+        "settings.max_examples={s.max_examples}, "
+        "but < 10% of examples satisfied assumptions"
+    )
+    max_shrinks = f"shrunk example {MAX_SHRINKS} times"
+    finished = "nothing left to do"
+    flaky = "test was flaky"
+    very_slow_shrinking = "shrinking was very slow"
+
+    def describe(self, settings):
+        return self.value.format(s=settings)
 
 
 class RunIsComplete(Exception):
@@ -76,19 +82,28 @@ class RunIsComplete(Exception):
 
 
 class ConjectureRunner:
-    def __init__(self, test_function, settings=None, random=None, database_key=None):
+    def __init__(
+        self,
+        test_function,
+        settings=None,
+        random=None,
+        database_key=None,
+        ignore_limits=False,
+    ):
         self._test_function = test_function
         self.settings = settings or Settings()
         self.shrinks = 0
+        self.finish_shrinking_deadline = None
         self.call_count = 0
-        self.event_call_counts = Counter()
         self.valid_examples = 0
         self.random = random or Random(getrandbits(128))
         self.database_key = database_key
-        self.status_runtimes = {}
+        self.ignore_limits = ignore_limits
 
-        self.all_drawtimes = []
-        self.all_runtimes = []
+        # Global dict of per-phase statistics, and a list of per-call stats
+        # which transfer to the global dict at the end of each phase.
+        self.statistics = {}
+        self.stats_per_test_case = []
 
         self.events_to_strings = WeakKeyDictionary()
 
@@ -122,6 +137,31 @@ class ConjectureRunner:
         # executed test case.
         self.__data_cache = LRUReusedCache(CACHE_SIZE)
 
+        # We ensure that the test has this much stack space remaining, no matter
+        # the size of the stack when called, to de-flake RecursionErrors (#2494).
+        self.__recursion_limit = sys.getrecursionlimit()
+        self.__pending_call_explanation = None
+
+    def explain_next_call_as(self, explanation):
+        self.__pending_call_explanation = explanation
+
+    def clear_call_explanation(self):
+        self.__pending_call_explanation = None
+
+    @contextmanager
+    def _log_phase_statistics(self, phase):
+        self.stats_per_test_case.clear()
+        start_time = time.perf_counter()
+        try:
+            yield
+        finally:
+            self.statistics[phase + "-phase"] = {
+                "duration-seconds": time.perf_counter() - start_time,
+                "test-cases": list(self.stats_per_test_case),
+                "distinct-failures": len(self.interesting_examples),
+                "shrinks-successful": self.shrinks,
+            }
+
     @property
     def should_optimise(self):
         return Phase.target in self.settings.phases
@@ -131,9 +171,21 @@ class ConjectureRunner:
 
     def __stoppable_test_function(self, data):
         """Run ``self._test_function``, but convert a ``StopTest`` exception
-        into a normal return.
+        into a normal return and avoid raising Flaky for RecursionErrors.
         """
+        depth = stack_depth_of_caller()
+        # Because we add to the recursion limit, to be good citizens we also add
+        # a check for unbounded recursion.  The default limit is 1000, so this can
+        # only ever trigger if something really strange is happening and it's hard
+        # to imagine an intentionally-deeply-recursive use of this code.
+        assert depth <= 1000, (
+            "Hypothesis would usually add %d to the stack depth of %d here, "
+            "but we are already much deeper than expected.  Aborting now, to "
+            "avoid extending the stack limit in an infinite loop..."
+            % (self.__recursion_limit, depth)
+        )
         try:
+            sys.setrecursionlimit(depth + self.__recursion_limit)
             self._test_function(data)
         except StopTest as e:
             if e.testcounter == data.testcounter:
@@ -145,8 +197,14 @@ class ConjectureRunner:
                 # need to re-raise it so that it will eventually reach the
                 # correct engine.
                 raise
+        finally:
+            sys.setrecursionlimit(self.__recursion_limit)
 
     def test_function(self, data):
+        if self.__pending_call_explanation is not None:
+            self.debug(self.__pending_call_explanation)
+            self.__pending_call_explanation = None
+
         assert isinstance(data.observer, TreeRecordingObserver)
         self.call_count += 1
 
@@ -164,7 +222,14 @@ class ConjectureRunner:
             # the KeyboardInterrupt, never continue to the code below.
             if not interrupted:  # pragma: no branch
                 data.freeze()
-                self.note_details(data)
+                call_stats = {
+                    "status": data.status.name.lower(),
+                    "runtime": data.finish_time - data.start_time,
+                    "drawtime": math.fsum(data.draw_times),
+                    "events": sorted({self.event_to_string(e) for e in data.events}),
+                }
+                self.stats_per_test_case.append(call_stats)
+                self.__data_cache[data.buffer] = data.as_result()
 
         self.debug_data(data)
 
@@ -221,6 +286,22 @@ class ConjectureRunner:
             if self.shrinks >= MAX_SHRINKS:
                 self.exit_with(ExitReason.max_shrinks)
 
+        if (
+            not self.ignore_limits
+            and self.finish_shrinking_deadline is not None
+            and self.finish_shrinking_deadline < time.perf_counter()
+        ):
+            # See https://github.com/HypothesisWorks/hypothesis/issues/2340
+            report(
+                "WARNING: Hypothesis has spent more than five minutes working to shrink "
+                "a failing example, and stopped because it is making very slow "
+                "progress.  When you re-run your tests, shrinking will resume and "
+                "may take this long before aborting again.\n"
+                "PLEASE REPORT THIS if you can provide a reproducing example, so that "
+                "we can improve shrinking performance for everyone."
+            )
+            self.exit_with(ExitReason.very_slow_shrinking)
+
         if not self.interesting_examples:
             # Note that this logic is reproduced to end the generation phase when
             # we have interesting examples.  Update that too if you change this!
@@ -362,15 +443,6 @@ class ConjectureRunner:
     def pareto_key(self):
         return self.sub_key(b"pareto")
 
-    def note_details(self, data):
-        self.__data_cache[data.buffer] = data.as_result()
-        runtime = max(data.finish_time - data.start_time, 0.0)
-        self.all_runtimes.append(runtime)
-        self.all_drawtimes.extend(data.draw_times)
-        self.status_runtimes.setdefault(data.status, []).append(runtime)
-        for event in set(map(self.event_to_string, data.events)):
-            self.event_call_counts[event] += 1
-
     def debug(self, message):
         if self.settings.verbosity >= Verbosity.debug:
             base_report(message)
@@ -408,11 +480,9 @@ class ConjectureRunner:
         status = repr(data.status)
 
         if data.status == Status.INTERESTING:
-            status = "%s (%r)" % (status, data.interesting_origin)
+            status = f"{status} ({data.interesting_origin!r})"
 
-        self.debug(
-            "%d bytes %r -> %s, %s" % (data.index, stack[0], status, data.output)
-        )
+        self.debug(f"{data.index} bytes {stack[0]!r} -> {status}, {data.output}")
 
     def run(self):
         with local_settings(self.settings):
@@ -464,7 +534,8 @@ class ConjectureRunner:
             corpus = sorted(
                 self.settings.database.fetch(self.database_key), key=sort_key
             )
-            desired_size = max(2, ceil(0.1 * self.settings.max_examples))
+            factor = 0.1 if (Phase.generate in self.settings.phases) else 1
+            desired_size = max(2, ceil(factor * self.settings.max_examples))
 
             if len(corpus) < desired_size:
                 extra_corpus = list(self.settings.database.fetch(self.secondary_key))
@@ -501,7 +572,12 @@ class ConjectureRunner:
                         break
 
     def exit_with(self, reason):
-        self.debug("exit_with(%s)" % (reason.name,))
+        if self.ignore_limits:
+            return
+        self.statistics["stopped-because"] = reason.describe(self.settings)
+        if self.best_observed_targets:
+            self.statistics["targets"] = dict(self.best_observed_targets)
+        self.debug(f"exit_with({reason.name})")
         self.exit_reason = reason
         raise RunIsComplete()
 
@@ -718,9 +794,9 @@ class ConjectureRunner:
 
                 group = self.random.choice(groups)
 
-                ex1, ex2 = [
+                ex1, ex2 = (
                     data.examples[i] for i in sorted(self.random.sample(group, 2))
-                ]
+                )
                 assert ex1.end <= ex2.start
 
                 replacements = [data.buffer[e.start : e.end] for e in [ex1, ex2]]
@@ -810,15 +886,17 @@ class ConjectureRunner:
             ParetoOptimiser(self).run()
 
     def _run(self):
-        self.reuse_existing_examples()
-        self.generate_new_examples()
-
-        # We normally run the targeting phase mixed in with the generate phase,
-        # but if we've been asked to run it but not generation then we have to
-        # run it explciitly on its own here.
-        if Phase.generate not in self.settings.phases:
-            self.optimise_targets()
-        self.shrink_interesting_examples()
+        with self._log_phase_statistics("reuse"):
+            self.reuse_existing_examples()
+        with self._log_phase_statistics("generate"):
+            self.generate_new_examples()
+            # We normally run the targeting phase mixed in with the generate phase,
+            # but if we've been asked to run it but not generation then we have to
+            # run it explciitly on its own here.
+            if Phase.generate not in self.settings.phases:
+                self.optimise_targets()
+        with self._log_phase_statistics("shrink"):
+            self.shrink_interesting_examples()
         self.exit_with(ExitReason.finished)
 
     def new_conjecture_data(self, prefix, max_length=BUFFER_SIZE, observer=None):
@@ -844,6 +922,12 @@ class ConjectureRunner:
 
         self.debug("Shrinking interesting examples")
 
+        # If the shrinking phase takes more than five minutes, abort it early and print
+        # a warning.   Many CI systems will kill a build after around ten minutes with
+        # no output, and appearing to hang isn't great for interactive use either -
+        # showing partially-shrunk examples is better than quitting with no examples!
+        self.finish_shrinking_deadline = time.perf_counter() + 300
+
         for prev_data in sorted(
             self.interesting_examples.values(), key=lambda d: sort_key(d.buffer)
         ):
@@ -864,7 +948,7 @@ class ConjectureRunner:
                 ),
                 key=lambda kv: (sort_key(kv[1].buffer), sort_key(repr(kv[0]))),
             )
-            self.debug("Shrinking %r" % (target,))
+            self.debug(f"Shrinking {target!r}")
 
             if not self.settings.report_multiple_bugs:
                 # If multi-bug reporting is disabled, we shrink our currently-minimal
@@ -907,13 +991,13 @@ class ConjectureRunner:
                     # of this reason for interestingness.
                     self.settings.database.delete(self.secondary_key, c)
 
-    def shrink(self, example, predicate):
-        s = self.new_shrinker(example, predicate)
+    def shrink(self, example, predicate=None, allow_transition=None):
+        s = self.new_shrinker(example, predicate, allow_transition)
         s.shrink()
         return s.shrink_target
 
-    def new_shrinker(self, example, predicate):
-        return Shrinker(self, example, predicate)
+    def new_shrinker(self, example, predicate=None, allow_transition=None):
+        return Shrinker(self, example, predicate, allow_transition)
 
     def cached_test_function(self, buffer, error_on_discard=False, extend=0):
         """Checks the tree to see if we've tested this buffer, and returns the
@@ -940,7 +1024,9 @@ class ConjectureRunner:
             return result
 
         try:
-            return check_result(self.__data_cache[buffer])
+            cached = check_result(self.__data_cache[buffer])
+            if cached.status > Status.OVERRUN or extend == 0:
+                return cached
         except KeyError:
             pass
 
@@ -980,11 +1066,12 @@ class ConjectureRunner:
         result = None
 
         data = self.new_conjecture_data(
-            prefix=max((buffer, dummy_data.buffer), key=len), max_length=max_length,
+            prefix=max((buffer, dummy_data.buffer), key=len), max_length=max_length
         )
         self.test_function(data)
         result = check_result(data.as_result())
-        self.__data_cache[buffer] = result
+        if extend == 0 or (result is not Overrun and len(result.buffer) <= len(buffer)):
+            self.__data_cache[buffer] = result
         return result
 
     def event_to_string(self, event):
