@@ -21,25 +21,19 @@ import sys
 from collections import OrderedDict, abc
 
 from hypothesis.errors import InvalidArgument
-from hypothesis.internal.compat import (
-    bit_length,
-    floor,
-    int_from_bytes,
-    qualname,
-    str_to_bytes,
-)
+from hypothesis.internal.compat import floor, int_from_bytes
 from hypothesis.internal.floats import int_to_float
 
 LABEL_MASK = 2 ** 64 - 1
 
 
 def calc_label_from_name(name: str) -> int:
-    hashed = hashlib.sha384(str_to_bytes(name)).digest()
+    hashed = hashlib.sha384(name.encode()).digest()
     return int_from_bytes(hashed[:8])
 
 
 def calc_label_from_cls(cls: type) -> int:
-    return calc_label_from_name(qualname(cls))
+    return calc_label_from_name(cls.__qualname__)
 
 
 def combine_labels(*labels: int) -> int:
@@ -53,15 +47,12 @@ def combine_labels(*labels: int) -> int:
 INTEGER_RANGE_DRAW_LABEL = calc_label_from_name("another draw in integer_range()")
 BIASED_COIN_LABEL = calc_label_from_name("biased_coin()")
 BIASED_COIN_INNER_LABEL = calc_label_from_name("inside biased_coin()")
-SAMPLE_IN_SAMPLER_LABLE = calc_label_from_name("a sample() in Sampler")
+SAMPLE_IN_SAMPLER_LABEL = calc_label_from_name("a sample() in Sampler")
 ONE_FROM_MANY_LABEL = calc_label_from_name("one more from many()")
 
-INT_SIZES = (8, 16, 32, 64, 128)
-INT_WEIGHTS = (4.0, 8.0, 1.0, 1.0, 0.5)
-
 
 def unbounded_integers(data):
-    size = INT_SIZES[Sampler(INT_WEIGHTS).sample(data)]
+    size = INT_SIZES[INT_SIZES_SAMPLER.sample(data)]
     r = data.draw_bits(size)
     sign = r & 1
     r >>= 1
@@ -89,7 +80,7 @@ def integer_range(data, lower, upper, center=None):
     elif center == lower:
         above = True
     else:
-        above = boolean(data)
+        above = not boolean(data)
 
     if above:
         gap = upper - center
@@ -98,14 +89,14 @@ def integer_range(data, lower, upper, center=None):
 
     assert gap > 0
 
-    bits = bit_length(gap)
+    bits = gap.bit_length()
     probe = gap + 1
 
     if bits > 24 and data.draw_bits(3):
         # For large ranges, we combine the uniform random distribution from draw_bits
         # with a weighting scheme with moderate chance.  Cutoff at 2 ** 24 so that our
         # choice of unicode characters is uniform but the 32bit distribution is not.
-        idx = Sampler(INT_WEIGHTS).sample(data)
+        idx = INT_SIZES_SAMPLER.sample(data)
         bits = min(bits, INT_SIZES[idx])
 
     while probe > gap:
@@ -126,28 +117,24 @@ def check_sample(values, strategy_name):
     if "numpy" in sys.modules and isinstance(values, sys.modules["numpy"].ndarray):
         if values.ndim != 1:
             raise InvalidArgument(
-                (
-                    "Only one-dimensional arrays are supported for sampling, "
-                    "and the given value has {ndim} dimensions (shape "
-                    "{shape}).  This array would give samples of array slices "
-                    "instead of elements!  Use np.ravel(values) to convert "
-                    "to a one-dimensional array, or tuple(values) if you "
-                    "want to sample slices."
-                ).format(ndim=values.ndim, shape=values.shape)
+                "Only one-dimensional arrays are supported for sampling, "
+                f"and the given value has {values.ndim} dimensions (shape "
+                f"{values.shape}).  This array would give samples of array slices "
+                "instead of elements!  Use np.ravel(values) to convert "
+                "to a one-dimensional array, or tuple(values) if you "
+                "want to sample slices."
             )
     elif not isinstance(values, (OrderedDict, abc.Sequence, enum.EnumMeta)):
         raise InvalidArgument(
-            "Cannot sample from {values}, not an ordered collection. "
-            "Hypothesis goes to some length to ensure that the {strategy} "
+            f"Cannot sample from {values!r}, not an ordered collection. "
+            f"Hypothesis goes to some length to ensure that the {strategy_name} "
             "strategy has stable results between runs. To replay a saved "
             "example, the sampled values must have the same iteration order "
             "on every run - ruling out sets, dicts, etc due to hash "
             "randomisation. Most cases can simply use `sorted(values)`, but "
             "mixed types or special values such as math.nan require careful "
             "handling - and note that when simplifying an example, "
-            "Hypothesis treats earlier values as simpler.".format(
-                values=repr(values), strategy=strategy_name
-            )
+            "Hypothesis treats earlier values as simpler."
         )
     if isinstance(values, range):
         return values
@@ -371,7 +358,7 @@ class Sampler:
         self.table.sort()
 
     def sample(self, data):
-        data.start_example(SAMPLE_IN_SAMPLER_LABLE)
+        data.start_example(SAMPLE_IN_SAMPLER_LABEL)
         i = integer_range(data, 0, len(self.table) - 1)
         base, alternate, alternate_chance = self.table[i]
         use_alternate = biased_coin(data, alternate_chance)
@@ -382,6 +369,10 @@ class Sampler:
             return base
 
 
+INT_SIZES = (8, 16, 32, 64, 128)
+INT_SIZES_SAMPLER = Sampler((4.0, 8.0, 1.0, 1.0, 0.5))
+
+
 class many:
     """Utility class for collections. Bundles up the logic we use for "should I
     keep drawing more values?" and handles starting and stopping examples in
