@@ -27,28 +27,13 @@ from tokenize import detect_encoding
 from types import ModuleType
 from typing import Callable, TypeVar
 
-from hypothesis.internal.compat import (
-    is_typed_named_tuple,
-    qualname,
-    str_to_bytes,
-    to_unicode,
-    update_code_location,
-)
+from hypothesis.internal.compat import is_typed_named_tuple, update_code_location
 from hypothesis.vendor.pretty import pretty
 
 C = TypeVar("C", bound=Callable)
 READTHEDOCS = os.environ.get("READTHEDOCS", None) == "True"
 
 
-def fully_qualified_name(f):
-    """Returns a unique identifier for f pointing to the module it was defined
-    on, and an containing functions."""
-    if f.__module__ is not None:
-        return f.__module__ + "." + qualname(f)
-    else:
-        return qualname(f)
-
-
 def is_mock(obj):
     """Determine if the given argument is a mock type."""
 
@@ -61,6 +46,13 @@ def is_mock(obj):
     return hasattr(obj, "hypothesis_internal_is_this_a_mock_check")
 
 
+def getfullargspec_except_self(target):
+    spec = inspect.getfullargspec(target)
+    if inspect.ismethod(target):
+        del spec.args[0]
+    return spec
+
+
 def function_digest(function):
     """Returns a string that is stable across multiple invocations across
     multiple processes and is prone to changing significantly in response to
@@ -70,19 +62,19 @@ def function_digest(function):
     """
     hasher = hashlib.sha384()
     try:
-        hasher.update(to_unicode(inspect.getsource(function)).encode("utf-8"))
+        hasher.update(inspect.getsource(function).encode())
     except (OSError, TypeError):
         pass
     try:
-        hasher.update(str_to_bytes(function.__name__))
+        hasher.update(function.__name__.encode())
     except AttributeError:
         pass
     try:
-        hasher.update(function.__module__.__name__.encode("utf-8"))
+        hasher.update(function.__module__.__name__.encode())
     except AttributeError:
         pass
     try:
-        hasher.update(str_to_bytes(repr(inspect.getfullargspec(function))))
+        hasher.update(repr(getfullargspec_except_self(function)).encode())
     except TypeError:
         pass
     try:
@@ -105,7 +97,7 @@ def required_args(target, args=(), kwargs=()):
     if inspect.isclass(target) and is_typed_named_tuple(target):
         provided = set(kwargs) | set(target._fields[: len(args)])
         return set(target._fields) - provided
-    # Then we try to do the right thing with inspect.getfullargspec
+    # Then we try to do the right thing with getfullargspec_except_self
     # Note that for classes we inspect the __init__ method, *unless* the class
     # has an explicit __signature__ attribute.  This allows us to support
     # runtime-generated constraints on **kwargs, as for e.g. Pydantic models.
@@ -136,7 +128,7 @@ def convert_keyword_arguments(function, args, kwargs):
     passed as positional and keyword args to the function. Unless function has
     kwonlyargs or **kwargs the dictionary will always be empty.
     """
-    argspec = inspect.getfullargspec(function)
+    argspec = getfullargspec_except_self(function)
     new_args = []
     kwargs = dict(kwargs)
 
@@ -171,8 +163,7 @@ def convert_keyword_arguments(function, args, kwargs):
         else:
             bad_kwarg = next(iter(kwargs))
             raise TypeError(
-                "%s() got an unexpected keyword argument %r"
-                % (function.__name__, bad_kwarg)
+                f"{function.__name__}() got an unexpected keyword argument {bad_kwarg!r}"
             )
     return tuple(new_args), kwargs
 
@@ -183,15 +174,14 @@ def convert_positional_arguments(function, args, kwargs):
 
     new_args will only be non-empty if function has a variadic argument.
     """
-    argspec = inspect.getfullargspec(function)
+    argspec = getfullargspec_except_self(function)
     new_kwargs = dict(argspec.kwonlydefaults or {})
     new_kwargs.update(kwargs)
     if not argspec.varkw:
         for k in new_kwargs.keys():
             if k not in argspec.args and k not in argspec.kwonlyargs:
                 raise TypeError(
-                    "%s() got an unexpected keyword argument %r"
-                    % (function.__name__, k)
+                    f"{function.__name__}() got an unexpected keyword argument {k!r}"
                 )
     if len(args) < len(argspec.args):
         for i in range(len(args), len(argspec.args) - len(argspec.defaults or ())):
@@ -203,9 +193,8 @@ def convert_positional_arguments(function, args, kwargs):
 
     if len(args) > len(argspec.args) and not argspec.varargs:
         raise TypeError(
-            "{}() takes at most {} positional arguments ({} given)".format(
-                function.__name__, len(argspec.args), len(args)
-            )
+            f"{function.__name__}() takes at most {len(argspec.args)} "
+            f"positional arguments ({len(args)} given)"
         )
 
     for arg, name in zip(args, argspec.args):
@@ -396,6 +385,9 @@ def get_pretty_function_description(f):
         # their module as __self__.  This might include c-extensions generally?
         if not (self is None or inspect.isclass(self) or inspect.ismodule(self)):
             return f"{self!r}.{name}"
+    elif isinstance(name, str) and getattr(dict, name, object()) is f:
+        # special case for keys/values views in from_type() / ghostwriter output
+        return f"dict.{name}"
     return name
 
 
@@ -405,23 +397,24 @@ def nicerepr(v):
     elif isinstance(v, type):
         return v.__name__
     else:
-        return pretty(v)
+        # With TypeVar T, show List[T] instead of TypeError on List[~T]
+        return re.sub(r"(\[)~([A-Z][a-z]*\])", r"\g<1>\g<2>", pretty(v))
 
 
 def arg_string(f, args, kwargs, reorder=True):
     if reorder:
         args, kwargs = convert_positional_arguments(f, args, kwargs)
 
-    argspec = inspect.getfullargspec(f)
+    argspec = getfullargspec_except_self(f)
 
     bits = []
 
     for a in argspec.args:
         if a in kwargs:
-            bits.append("{}={}".format(a, nicerepr(kwargs.pop(a))))
+            bits.append(f"{a}={nicerepr(kwargs.pop(a))}")
     if kwargs:
         for a in sorted(kwargs):
-            bits.append("{}={}".format(a, nicerepr(kwargs[a])))
+            bits.append(f"{a}={nicerepr(kwargs[a])}")
 
     return ", ".join([nicerepr(x) for x in args] + bits)
 
@@ -446,10 +439,8 @@ def source_exec_as_module(source):
     except KeyError:
         pass
 
-    result = ModuleType(
-        "hypothesis_temporary_module_%s"
-        % (hashlib.sha384(str_to_bytes(source)).hexdigest(),)
-    )
+    hexdigest = hashlib.sha384(source.encode()).hexdigest()
+    result = ModuleType("hypothesis_temporary_module_" + hexdigest)
     assert isinstance(source, str)
     exec(source, result.__dict__)
     eval_cache[source] = result
@@ -492,7 +483,7 @@ def define_function_signature(name, docstring, argspec):
         check_valid_identifier(a)
 
     def accept(f):
-        fargspec = inspect.getfullargspec(f)
+        fargspec = getfullargspec_except_self(f)
         must_pass_as_kwargs = []
         invocation_parts = []
         for a in argspec.args:
@@ -576,7 +567,7 @@ def proxies(target):
                 define_function_signature(
                     target.__name__.replace("<lambda>", "_lambda_"),
                     target.__doc__,
-                    inspect.getfullargspec(target),
+                    getfullargspec_except_self(target),
                 )(proxy)
             )
         )
