@@ -26,12 +26,13 @@ definitions it links to.  If not, report the bug!
 import os.path
 import string
 
+from hypothesis import strategies as st
 from hypothesis.errors import InvalidArgument
 from hypothesis.internal.conjecture import utils as cu
-from hypothesis.strategies._internal import core as st
-from hypothesis.strategies._internal.strategies import SearchStrategy
+from hypothesis.strategies._internal.utils import defines_strategy
 
 URL_SAFE_CHARACTERS = frozenset(string.ascii_letters + string.digits + "$-_.+!*'(),~")
+FRAGMENT_SAFE_CHARACTERS = URL_SAFE_CHARACTERS | {"?", "/"}
 
 
 # This file is sourced from http://data.iana.org/TLD/tlds-alpha-by-domain.txt
@@ -50,20 +51,18 @@ assert _tlds[0].startswith("#")
 TOP_LEVEL_DOMAINS = ["COM"] + sorted(_tlds[1:], key=len)
 
 
-class DomainNameStrategy(SearchStrategy):
+class DomainNameStrategy(st.SearchStrategy):
     @staticmethod
     def clean_inputs(minimum, maximum, value, variable_name):
         if value is None:
             value = maximum
         elif not isinstance(value, int):
             raise InvalidArgument(
-                "Expected integer but %s is a %s"
-                % (variable_name, type(value).__name__)
+                f"Expected integer but {variable_name} is a {type(value).__name__}"
             )
         elif not minimum <= value <= maximum:
             raise InvalidArgument(
-                "Invalid value %r < %s=%r < %r"
-                % (minimum, variable_name, value, maximum)
+                f"Invalid value {minimum!r} < {variable_name}={value!r} < {maximum!r}"
             )
         return value
 
@@ -111,7 +110,7 @@ class DomainNameStrategy(SearchStrategy):
             .filter(lambda tld: len(tld) + 2 <= self.max_length)
             .flatmap(
                 lambda tld: st.tuples(
-                    *[st.sampled_from([c.lower(), c.upper()]) for c in tld]
+                    *(st.sampled_from([c.lower(), c.upper()]) for c in tld)
                 ).map("".join)
             )
         )
@@ -131,18 +130,37 @@ class DomainNameStrategy(SearchStrategy):
         return domain
 
 
-@st.defines_strategy(force_reusable_values=True)
+@defines_strategy(force_reusable_values=True)
 def domains(
     *, max_length: int = 255, max_element_length: int = 63
-) -> SearchStrategy[str]:
+) -> st.SearchStrategy[str]:
     """Generate :rfc:`1035` compliant fully qualified domain names."""
     return DomainNameStrategy(
         max_length=max_length, max_element_length=max_element_length
     )
 
 
-@st.defines_strategy(force_reusable_values=True)
-def urls() -> SearchStrategy[str]:
+# The `urls()` strategy uses this to generate URL fragments (e.g. "#foo").
+# It has been extracted to top-level so that we can test it independently
+# of `urls()`, which helps with getting non-flaky coverage of the lambda.
+_url_fragments_strategy = (
+    st.lists(
+        st.builds(
+            lambda char, encode: f"%{ord(char):02X}"
+            if (encode or char not in FRAGMENT_SAFE_CHARACTERS)
+            else char,
+            st.characters(min_codepoint=0, max_codepoint=255),
+            st.booleans(),
+        ),
+        min_size=1,
+    )
+    .map("".join)
+    .map("#{}".format)
+)
+
+
+@defines_strategy(force_reusable_values=True)
+def urls() -> st.SearchStrategy[str]:
     """A strategy for :rfc:`3986`, generating http/https URLs."""
 
     def url_encode(s):
@@ -153,5 +171,10 @@ def urls() -> SearchStrategy[str]:
     paths = st.lists(st.text(string.printable).map(url_encode)).map("/".join)
 
     return st.builds(
-        "{}://{}{}/{}".format, schemes, domains(), st.just("") | ports, paths
+        "{}://{}{}/{}{}".format,
+        schemes,
+        domains(),
+        st.just("") | ports,
+        paths,
+        st.just("") | _url_fragments_strategy,
     )
