@@ -24,7 +24,7 @@ import typing
 from decimal import Context, Decimal, localcontext
 from fractions import Fraction
 from functools import reduce
-from inspect import Parameter, getfullargspec, isabstract, isclass, signature
+from inspect import Parameter, Signature, getfullargspec, isabstract, isclass, signature
 from types import FunctionType
 from typing import (
     Any,
@@ -95,7 +95,12 @@ from hypothesis.strategies._internal.deferred import DeferredStrategy
 from hypothesis.strategies._internal.functions import FunctionStrategy
 from hypothesis.strategies._internal.lazy import LazyStrategy
 from hypothesis.strategies._internal.misc import just, none, nothing
-from hypothesis.strategies._internal.numbers import Real, floats, integers
+from hypothesis.strategies._internal.numbers import (
+    IntegersStrategy,
+    Real,
+    floats,
+    integers,
+)
 from hypothesis.strategies._internal.recursive import RecursiveStrategy
 from hypothesis.strategies._internal.shared import SharedStrategy
 from hypothesis.strategies._internal.strategies import (
@@ -111,6 +116,11 @@ from hypothesis.strategies._internal.strings import (
 from hypothesis.strategies._internal.utils import cacheable, defines_strategy
 from hypothesis.utils.conventions import InferType, infer, not_set
 
+try:
+    from typing import Protocol
+except ImportError:  # < py3.8
+    Protocol = object  # type: ignore[assignment]
+
 UniqueBy = Union[Callable[[Ex], Hashable], Tuple[Callable[[Ex], Hashable], ...]]
 
 
@@ -158,6 +168,17 @@ def sampled_from(elements):
     """
     values = check_sample(elements, "sampled_from")
     if not values:
+        if (
+            isinstance(elements, type)
+            and issubclass(elements, enum.Enum)
+            and vars(elements).get("__annotations__")
+        ):
+            # See https://github.com/HypothesisWorks/hypothesis/issues/2923
+            raise InvalidArgument(
+                f"Cannot sample from {elements.__module__}.{elements.__name__} "
+                "because it contains no elements.  It does however have annotations, "
+                "so maybe you tried to write an enum as if it was a dataclass?"
+            )
         raise InvalidArgument("Cannot sample from a length-zero sequence.")
     if len(values) == 1:
         return just(values[0])
@@ -254,7 +275,7 @@ def lists(
             and len(unique_by) == 1
             and (
                 # Introspection for either `itemgetter(0)`, or `lambda x: x[0]`
-                isinstance(unique_by[0], operator.itemgetter)  # type: ignore
+                isinstance(unique_by[0], operator.itemgetter)
                 and repr(unique_by[0]) == "operator.itemgetter(0)"
                 or isinstance(unique_by[0], FunctionType)
                 and re.fullmatch(
@@ -267,6 +288,20 @@ def lists(
             tuple_suffixes = TupleStrategy(elements.element_strategies[1:])
             elements = elements.element_strategies[0]
 
+        # UniqueSampledListStrategy offers a substantial performance improvement for
+        # unique arrays with few possible elements, e.g. of eight-bit integer types.
+        if (
+            isinstance(elements, IntegersStrategy)
+            and None not in (elements.start, elements.end)
+            and (elements.end - elements.start) <= 255
+        ):
+            elements = SampledFromStrategy(
+                sorted(range(elements.start, elements.end + 1), key=abs)
+                if elements.end < 0 or elements.start > 0
+                else list(range(0, elements.end + 1))
+                + list(range(-1, elements.start - 1, -1))
+            )
+
         if isinstance(elements, SampledFromStrategy):
             element_count = len(elements.elements)
             if min_size > element_count:
@@ -385,13 +420,12 @@ def fixed_dictionaries(
     """Generates a dictionary of the same type as mapping with a fixed set of
     keys mapping to strategies. ``mapping`` must be a dict subclass.
 
-    Generated values have all keys present in mapping, with the
-    corresponding values drawn from mapping[key]. If mapping is an
-    instance of OrderedDict the keys will also be in the same order,
-    otherwise the order is arbitrary.
+    Generated values have all keys present in mapping, in iteration order,
+    with the corresponding values drawn from mapping[key].
 
     If ``optional`` is passed, the generated value *may or may not* contain each
     key from ``optional`` and a value drawn from the corresponding strategy.
+    Generated values may contain optional keys in an arbitrary order.
 
     Examples from this strategy shrink by shrinking each individual value in
     the generated dictionary, and omitting optional key-value pairs.
@@ -411,7 +445,7 @@ def fixed_dictionaries(
         if set(mapping) & set(optional):
             raise InvalidArgument(
                 "The following keys were in both mapping and optional, "
-                "which is invalid: %r" % (set(mapping) & set(optional))
+                f"which is invalid: {set(mapping) & set(optional)!r}"
             )
         return FixedAndOptionalKeysDictStrategy(mapping, optional)
     return FixedKeysDictStrategy(mapping)
@@ -520,9 +554,9 @@ def characters(
     overlap = set(blacklist_characters).intersection(whitelist_characters)
     if overlap:
         raise InvalidArgument(
-            "Characters %r are present in both whitelist_characters=%r, and "
-            "blacklist_characters=%r"
-            % (sorted(overlap), whitelist_characters, blacklist_characters)
+            f"Characters {sorted(overlap)!r} are present in both "
+            f"whitelist_characters={whitelist_characters!r}, and "
+            f"blacklist_characters={blacklist_characters!r}"
         )
     blacklist_categories = as_general_categories(
         blacklist_categories, "blacklist_categories"
@@ -543,9 +577,9 @@ def characters(
     both_cats = set(blacklist_categories or ()).intersection(whitelist_categories or ())
     if both_cats:
         raise InvalidArgument(
-            "Categories %r are present in both whitelist_categories=%r, and "
-            "blacklist_categories=%r"
-            % (sorted(both_cats), whitelist_categories, blacklist_categories)
+            f"Categories {sorted(both_cats)!r} are present in both "
+            f"whitelist_categories={whitelist_categories!r}, and "
+            f"blacklist_categories={blacklist_categories!r}"
         )
 
     return OneCharStringStrategy(
@@ -594,14 +628,13 @@ def text(
         if non_string:
             raise InvalidArgument(
                 "The following elements in alphabet are not unicode "
-                "strings:  %r" % (non_string,)
+                f"strings:  {non_string!r}"
             )
         not_one_char = [c for c in alphabet if len(c) != 1]
         if not_one_char:
             raise InvalidArgument(
-                "The following elements in alphabet are not of length "
-                "one, which leads to violation of size constraints:  %r"
-                % (not_one_char,)
+                "The following elements in alphabet are not of length one, "
+                f"which leads to violation of size constraints:  {not_one_char!r}"
             )
         char_strategy = (
             characters(whitelist_categories=(), whitelist_characters=alphabet)
@@ -752,7 +785,7 @@ class BuildsStrategy(SearchStrategy):
     def do_draw(self, data):
         try:
             return self.target(
-                *[data.draw(a) for a in self.args],
+                *(data.draw(a) for a in self.args),
                 **{k: data.draw(v) for k, v in self.kwargs.items()},
             )
         except TypeError as err:
@@ -766,6 +799,15 @@ class BuildsStrategy(SearchStrategy):
                     f"Calling {name} with no arguments raised an error - "
                     f"try using sampled_from({name}) instead of builds({name})"
                 ) from err
+            if not (self.args or self.kwargs):
+                from .types import is_a_new_type, is_generic_type
+
+                if is_a_new_type(self.target) or is_generic_type(self.target):
+                    raise InvalidArgument(
+                        f"Calling {self.target!r} with no arguments raised an "
+                        f"error - try using from_type({self.target!r}) instead "
+                        f"of builds({self.target!r})"
+                    ) from err
             raise
 
     def validate(self):
@@ -839,9 +881,6 @@ def builds(
             )
         for kw in set(hints) & (required | to_infer):
             kwargs[kw] = from_type(hints[kw])
-
-    # Mypy doesn't realise that `infer` is gone from kwargs now
-    # and thinks that target and args have the same (union) type.
     return BuildsStrategy(target, args, kwargs)
 
 
@@ -864,6 +903,7 @@ if sys.version_info[:2] >= (3, 8):  # pragma: no cover
 
 
 @cacheable
+@defines_strategy(never_lazy=True)
 def from_type(thing: Type[Ex]) -> SearchStrategy[Ex]:
     """Looks up the appropriate search strategy for the given type.
 
@@ -955,8 +995,8 @@ def _from_type(thing: Type[Ex]) -> SearchStrategy[Ex]:
             strategy = strat_or_callable
         if not isinstance(strategy, SearchStrategy):
             raise ResolutionFailed(
-                "Error: %s was registered for %r, but returned non-strategy %r"
-                % (thing, nicerepr(strat_or_callable), strategy)
+                f"Error: {thing} was registered for {nicerepr(strat_or_callable)}, "
+                f"but returned non-strategy {strategy!r}"
             )
         if strategy.is_empty:
             raise ResolutionFailed(f"Error: {thing!r} resolved to an empty strategy")
@@ -990,7 +1030,15 @@ def _from_type(thing: Type[Ex]) -> SearchStrategy[Ex]:
                 else:
                     literals.append(arg)
             return sampled_from(literals)
-        raise InvalidArgument(f"thing={thing} must be a type")
+        if isinstance(thing, str):
+            # See https://github.com/HypothesisWorks/hypothesis/issues/3016
+            raise InvalidArgument(
+                f"Got {thing!r} as a type annotation, but the forward-reference "
+                "could not be resolved from a string to a type.  Consider using "
+                "`from __future__ import annotations` instead of forward-reference "
+                "strings."
+            )
+        raise InvalidArgument(f"thing={thing!r} must be a type")  # pragma: no cover
     # Now that we know `thing` is a type, the first step is to check for an
     # explicitly registered strategy.  This is the best (and hopefully most
     # common) way to resolve a type to a strategy.  Note that the value in the
@@ -1052,17 +1100,7 @@ def _from_type(thing: Type[Ex]) -> SearchStrategy[Ex]:
     # may be able to fall back on type annotations.
     if issubclass(thing, enum.Enum):
         return sampled_from(thing)
-    # If we know that builds(thing) will fail, give a better error message
-    required = required_args(thing)
-    if required and not (
-        required.issubset(get_type_hints(thing))
-        or attr.has(thing)
-        or is_typed_named_tuple(thing)  # weird enough that we have a specific check
-    ):
-        raise ResolutionFailed(
-            f"Could not resolve {thing!r} to a strategy; consider "
-            "using register_type_strategy"
-        )
+
     # Finally, try to build an instance by calling the type object.  Unlike builds(),
     # this block *does* try to infer strategies for arguments with default values.
     # That's because of the semantic different; builds() -> "call this with ..."
@@ -1070,6 +1108,17 @@ def _from_type(thing: Type[Ex]) -> SearchStrategy[Ex]:
     # me arbitrary instances" so the greater variety is acceptable.
     # And if it's *too* varied, express your opinions with register_type_strategy()
     if not isabstract(thing):
+        # If we know that builds(thing) will fail, give a better error message
+        required = required_args(thing)
+        if required and not (
+            required.issubset(get_type_hints(thing))
+            or attr.has(thing)
+            or is_typed_named_tuple(thing)  # weird enough that we have a specific check
+        ):
+            raise ResolutionFailed(
+                f"Could not resolve {thing!r} to a strategy; consider "
+                "using register_type_strategy"
+            )
         try:
             hints = get_type_hints(thing)
             params = signature(thing).parameters
@@ -1092,7 +1141,17 @@ def _from_type(thing: Type[Ex]) -> SearchStrategy[Ex]:
             f"Could not resolve {thing!r} to a strategy, because it is an abstract "
             "type without any subclasses. Consider using register_type_strategy"
         )
-    return sampled_from(subclasses).flatmap(from_type)
+    subclass_strategies = nothing()
+    for sc in subclasses:
+        try:
+            subclass_strategies |= _from_type(sc)
+        except Exception:
+            pass
+    if subclass_strategies.is_empty:
+        # We're unable to resolve subclasses now, but we might be able to later -
+        # so we'll just go back to the mixed distribution.
+        return sampled_from(subclasses).flatmap(from_type)
+    return subclass_strategies
 
 
 @cacheable
@@ -1128,16 +1187,16 @@ def fractions(
 
     if max_denominator is not None:
         if max_denominator < 1:
-            raise InvalidArgument("max_denominator=%r must be >= 1" % max_denominator)
+            raise InvalidArgument(f"max_denominator={max_denominator!r} must be >= 1")
         if min_value is not None and min_value.denominator > max_denominator:
             raise InvalidArgument(
-                "The min_value=%r has a denominator greater than the "
-                "max_denominator=%r" % (min_value, max_denominator)
+                f"The min_value={min_value!r} has a denominator greater than the "
+                f"max_denominator={max_denominator!r}"
             )
         if max_value is not None and max_value.denominator > max_denominator:
             raise InvalidArgument(
-                "The max_value=%r has a denominator greater than the "
-                "max_denominator=%r" % (max_value, max_denominator)
+                f"The max_value={max_value!r} has a denominator greater than the "
+                f"max_denominator={max_denominator!r}"
             )
 
     if min_value is not None and min_value == max_value:
@@ -1237,7 +1296,7 @@ def decimals(
     # Convert min_value and max_value to Decimal values, and validate args
     check_valid_integer(places, "places")
     if places is not None and places < 0:
-        raise InvalidArgument("places=%r may not be negative" % places)
+        raise InvalidArgument(f"places={places!r} may not be negative")
     min_value = _as_finite_decimal(min_value, "min_value", allow_infinity)
     max_value = _as_finite_decimal(max_value, "max_value", allow_infinity)
     check_valid_interval(min_value, max_value, "min_value", "max_value")
@@ -1293,6 +1352,7 @@ def decimals(
     return strat | (sampled_from(special) if special else nothing())
 
 
+@defines_strategy(never_lazy=True)
 def recursive(
     base: SearchStrategy[Ex],
     extend: Callable[[SearchStrategy[Any]], SearchStrategy[T]],
@@ -1367,6 +1427,33 @@ class CompositeStrategy(SearchStrategy):
         return calc_label_from_cls(self.definition)
 
 
+class DrawFn(Protocol):
+    """This type only exists so that you can write type hints for functions
+    decorated with :func:`@composite <hypothesis.strategies.composite>`.
+
+    .. code-block:: python
+
+        @composite
+        def list_and_index(draw: DrawFn) -> Tuple[int, str]:
+            i = draw(integers())  # type inferred as 'int'
+            s = draw(text())  # type inferred as 'str'
+
+    """
+
+    def __init__(self):
+        raise TypeError("Protocols cannot be instantiated")  # pragma: no cover
+
+    # On Python 3.8+, Protocol overrides our signature for __init__,
+    # so we override it right back to make the docs look nice.
+    __signature__: Signature = Signature(parameters=[])
+
+    # We define this as a callback protocol because a simple typing.Callable is
+    # insufficient to fully represent the interface, due to the optional `label`
+    # parameter.
+    def __call__(self, strategy: SearchStrategy[Ex], label: object = None) -> Ex:
+        raise NotImplementedError
+
+
 @cacheable
 def composite(f: Callable[..., Ex]) -> Callable[..., SearchStrategy[Ex]]:
     """Defines a strategy that is built out of potentially arbitrarily many
@@ -1455,15 +1542,15 @@ def complex_numbers(
         allow_infinity = bool(max_magnitude is None)
     elif allow_infinity and max_magnitude is not None:
         raise InvalidArgument(
-            "Cannot have allow_infinity=%r with max_magnitude=%r"
-            % (allow_infinity, max_magnitude)
+            f"Cannot have allow_infinity={allow_infinity!r} with "
+            f"max_magnitude={max_magnitude!r}"
         )
     if allow_nan is None:
         allow_nan = bool(min_magnitude == 0 and max_magnitude is None)
     elif allow_nan and not (min_magnitude == 0 and max_magnitude is None):
         raise InvalidArgument(
-            "Cannot have allow_nan=%r, min_magnitude=%r max_magnitude=%r"
-            % (allow_nan, min_magnitude, max_magnitude)
+            f"Cannot have allow_nan={allow_nan!r}, min_magnitude={min_magnitude!r} "
+            f"max_magnitude={max_magnitude!r}"
         )
     allow_kw = {"allow_nan": allow_nan, "allow_infinity": allow_infinity}
 
@@ -1498,6 +1585,7 @@ def complex_numbers(
     return constrained_complex()
 
 
+@defines_strategy(never_lazy=True)
 def shared(
     base: SearchStrategy[Ex],
     *,
@@ -1632,12 +1720,13 @@ class DataStrategy(SearchStrategy):
 
     def __not_a_first_class_strategy(self, name):
         raise InvalidArgument(
-            "Cannot call %s on a DataStrategy. You should probably be using "
-            "@composite for whatever it is you're trying to do." % (name,)
+            f"Cannot call {name} on a DataStrategy. You should probably "
+            "be using @composite for whatever it is you're trying to do."
         )
 
 
 @cacheable
+@defines_strategy(never_lazy=True)
 def data() -> SearchStrategy[DataObject]:
     """This isn't really a normal strategy, but instead gives you an object
     which can be used to draw data interactively from other strategies.
@@ -1680,7 +1769,7 @@ def register_type_strategy(
     from hypothesis.strategies._internal import types
 
     if not types.is_a_type(custom_type):
-        raise InvalidArgument("custom_type=%r must be a type")
+        raise InvalidArgument(f"custom_type={custom_type!r} must be a type")
     elif not (isinstance(strategy, SearchStrategy) or callable(strategy)):
         raise InvalidArgument(
             "strategy=%r must be a SearchStrategy, or a function that takes "
@@ -1702,6 +1791,7 @@ def register_type_strategy(
 
 
 @cacheable
+@defines_strategy(never_lazy=True)
 def deferred(definition: Callable[[], SearchStrategy[Ex]]) -> SearchStrategy[Ex]:
     """A deferred strategy allows you to write a strategy that references other
     strategies that have not yet been defined. This allows for the easy
@@ -1786,7 +1876,7 @@ def functions(
     if not callable(like):
         raise InvalidArgument(
             "The first argument to functions() must be a callable to imitate, "
-            "but got non-callable like=%r" % (nicerepr(like),)
+            f"but got non-callable like={nicerepr(like)!r}"
         )
 
     if returns is None:
@@ -1811,13 +1901,9 @@ def slices(draw: Any, size: int) -> slice:
     if size == 0:
         step = draw(none() | integers().filter(bool))
         return slice(None, None, step)
-
-    min_start = min_stop = 0
-    max_start = max_stop = size
-    min_step = 1
     # For slices start is inclusive and stop is exclusive
-    start = draw(integers(min_start, max_start) | none())
-    stop = draw(integers(min_stop, max_stop) | none())
+    start = draw(integers(0, size - 1) | none())
+    stop = draw(integers(0, size) | none())
 
     # Limit step size to be reasonable
     if start is None and stop is None:
@@ -1829,14 +1915,16 @@ def slices(draw: Any, size: int) -> slice:
     else:
         max_step = abs(start - stop)
 
-    step = draw(integers(min_step, max_step or 1))
+    step = draw(integers(1, max_step or 1))
 
-    if (stop or 0) < (start or 0):
+    if (draw(booleans()) and start == stop) or (stop or 0) < (start or 0):
         step *= -1
 
     if draw(booleans()) and start is not None:
         start -= size
     if draw(booleans()) and stop is not None:
         stop -= size
+    if (not draw(booleans())) and step == 1:
+        step = None
 
     return slice(start, stop, step)
