@@ -1,7 +1,7 @@
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis/
 #
-# Most of this work is copyright (C) 2013-2020 David R. MacIver
+# Most of this work is copyright (C) 2013-2021 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -14,19 +14,38 @@
 # END HEADER
 
 import datetime as dt
+import os.path
 from calendar import monthrange
+from functools import lru_cache
 from typing import Optional
 
 from hypothesis.errors import InvalidArgument
 from hypothesis.internal.conjecture import utils
 from hypothesis.internal.validation import check_type, check_valid_interval
-from hypothesis.strategies._internal.core import (
-    defines_strategy_with_reusable_values,
-    deprecated_posargs,
-    just,
-    none,
-)
+from hypothesis.strategies._internal.core import sampled_from
+from hypothesis.strategies._internal.misc import just, none
 from hypothesis.strategies._internal.strategies import SearchStrategy
+from hypothesis.strategies._internal.utils import defines_strategy
+
+# These standard-library modules are required for the timezones() and
+# timezone_keys() strategies, but not present in older versions of Python.
+# We therefore try to import them here, but only raise errors recommending
+# `pip install hypothesis[zoneinfo]` to install the backports (if needed)
+# when those strategies are actually used.
+try:
+    import importlib.resources as importlib_resources
+except ImportError:
+    try:
+        import importlib_resources  # type: ignore
+    except ImportError:
+        importlib_resources = None  # type: ignore
+try:
+    import zoneinfo
+except ImportError:
+    try:
+        from backports import zoneinfo  # type: ignore
+    except ImportError:
+        zoneinfo = None
 
 DATENAMES = ("year", "month", "day")
 TIMENAMES = ("hour", "minute", "second", "microsecond")
@@ -42,7 +61,7 @@ def is_pytz_timezone(tz):
 def replace_tzinfo(value, timezone):
     if is_pytz_timezone(timezone):
         # Pytz timezones are a little complicated, and using the .replace method
-        # can cause some wierd issues, so we use their special "localise" instead.
+        # can cause some weird issues, so we use their special "localize" instead.
         #
         # We use the fold attribute as a convenient boolean for is_dst, even though
         # they're semantically distinct.  For ambiguous or imaginary hours, fold says
@@ -54,16 +73,20 @@ def replace_tzinfo(value, timezone):
         # WARNING: this is INCORRECT for timezones with negative DST offsets such as
         #       "Europe/Dublin", but it's unclear what we could do instead beyond
         #       documenting the problem and recommending use of `dateutil` instead.
-        #
-        # TODO: after dropping Python 3.5 support we won't need the getattr
-        return timezone.localize(value, is_dst=not getattr(value, "fold", 0))
+        return timezone.localize(value, is_dst=not value.fold)
     return value.replace(tzinfo=timezone)
 
 
 def datetime_does_not_exist(value):
-    # This function tests whether the given datetime can be round-tripped to and
-    # from UTC.  It is an exact inverse of (and very similar to) the dateutil method
-    # https://dateutil.readthedocs.io/en/stable/tz.html#dateutil.tz.datetime_exists
+    """This function tests whether the given datetime can be round-tripped to and
+    from UTC.  It is an exact inverse of (and very similar to) the dateutil method
+    https://dateutil.readthedocs.io/en/stable/tz.html#dateutil.tz.datetime_exists
+    """
+    # Naive datetimes cannot be imaginary, but we need this special case because
+    # chaining .astimezone() ends with *the system local timezone*, not None.
+    # See bug report in https://github.com/HypothesisWorks/hypothesis/issues/2662
+    if value.tzinfo is None:
+        return False
     try:
         # Does the naive portion of the datetime change when round-tripped to
         # UTC?  If so, or if this overflows, we say that it does not exist.
@@ -78,18 +101,15 @@ def datetime_does_not_exist(value):
     return value != roundtrip
 
 
-def draw_capped_multipart(data, min_value, max_value):
+def draw_capped_multipart(
+    data, min_value, max_value, duration_names=DATENAMES + TIMENAMES
+):
     assert isinstance(min_value, (dt.date, dt.time, dt.datetime))
     assert type(min_value) == type(max_value)
     assert min_value <= max_value
     result = {}
     cap_low, cap_high = True, True
-    duration_names_by_type = {
-        dt.date: DATENAMES,
-        dt.time: TIMENAMES,
-        dt.datetime: DATENAMES + TIMENAMES,
-    }
-    for name in duration_names_by_type[type(min_value)]:
+    for name in duration_names:
         low = getattr(min_value if cap_low else dt.datetime.min, name)
         high = getattr(max_value if cap_high else dt.datetime.max, name)
         if name == "day" and not cap_high:
@@ -151,14 +171,13 @@ class DatetimeStrategy(SearchStrategy):
             data.mark_invalid()
 
 
-@defines_strategy_with_reusable_values
-@deprecated_posargs
+@defines_strategy(force_reusable_values=True)
 def datetimes(
     min_value: dt.datetime = dt.datetime.min,
     max_value: dt.datetime = dt.datetime.max,
     *,
     timezones: SearchStrategy[Optional[dt.tzinfo]] = none(),
-    allow_imaginary: bool = True
+    allow_imaginary: bool = True,
 ) -> SearchStrategy[dt.datetime]:
     """datetimes(min_value=datetime.datetime.min, max_value=datetime.datetime.max, *, timezones=none(), allow_imaginary=True)
 
@@ -197,14 +216,14 @@ def datetimes(
     check_type(dt.datetime, min_value, "min_value")
     check_type(dt.datetime, max_value, "max_value")
     if min_value.tzinfo is not None:
-        raise InvalidArgument("min_value=%r must not have tzinfo" % (min_value,))
+        raise InvalidArgument(f"min_value={min_value!r} must not have tzinfo")
     if max_value.tzinfo is not None:
-        raise InvalidArgument("max_value=%r must not have tzinfo" % (max_value,))
+        raise InvalidArgument(f"max_value={max_value!r} must not have tzinfo")
     check_valid_interval(min_value, max_value, "min_value", "max_value")
     if not isinstance(timezones, SearchStrategy):
         raise InvalidArgument(
-            "timezones=%r must be a SearchStrategy that can provide tzinfo "
-            "for datetimes (either None or dt.tzinfo objects)" % (timezones,)
+            f"timezones={timezones!r} must be a SearchStrategy that can "
+            "provide tzinfo for datetimes (either None or dt.tzinfo objects)"
         )
     return DatetimeStrategy(min_value, max_value, timezones, allow_imaginary)
 
@@ -216,18 +235,17 @@ class TimeStrategy(SearchStrategy):
         self.tz_strat = timezones_strat
 
     def do_draw(self, data):
-        result = draw_capped_multipart(data, self.min_value, self.max_value)
+        result = draw_capped_multipart(data, self.min_value, self.max_value, TIMENAMES)
         tz = data.draw(self.tz_strat)
         return dt.time(**result, tzinfo=tz)
 
 
-@defines_strategy_with_reusable_values
-@deprecated_posargs
+@defines_strategy(force_reusable_values=True)
 def times(
     min_value: dt.time = dt.time.min,
     max_value: dt.time = dt.time.max,
     *,
-    timezones: SearchStrategy[Optional[dt.tzinfo]] = none()
+    timezones: SearchStrategy[Optional[dt.tzinfo]] = none(),
 ) -> SearchStrategy[dt.time]:
     """times(min_value=datetime.time.min, max_value=datetime.time.max, *, timezones=none())
 
@@ -241,9 +259,9 @@ def times(
     check_type(dt.time, min_value, "min_value")
     check_type(dt.time, max_value, "max_value")
     if min_value.tzinfo is not None:
-        raise InvalidArgument("min_value=%r must not have tzinfo" % min_value)
+        raise InvalidArgument(f"min_value={min_value!r} must not have tzinfo")
     if max_value.tzinfo is not None:
-        raise InvalidArgument("max_value=%r must not have tzinfo" % max_value)
+        raise InvalidArgument(f"max_value={max_value!r} must not have tzinfo")
     check_valid_interval(min_value, max_value, "min_value", "max_value")
     return TimeStrategy(min_value, max_value, timezones)
 
@@ -257,10 +275,12 @@ class DateStrategy(SearchStrategy):
         self.max_value = max_value
 
     def do_draw(self, data):
-        return dt.date(**draw_capped_multipart(data, self.min_value, self.max_value))
+        return dt.date(
+            **draw_capped_multipart(data, self.min_value, self.max_value, DATENAMES)
+        )
 
 
-@defines_strategy_with_reusable_values
+@defines_strategy(force_reusable_values=True)
 def dates(
     min_value: dt.date = dt.date.min, max_value: dt.date = dt.date.max
 ) -> SearchStrategy[dt.date]:
@@ -300,7 +320,7 @@ class TimedeltaStrategy(SearchStrategy):
         return dt.timedelta(**result)
 
 
-@defines_strategy_with_reusable_values
+@defines_strategy(force_reusable_values=True)
 def timedeltas(
     min_value: dt.timedelta = dt.timedelta.min,
     max_value: dt.timedelta = dt.timedelta.max,
@@ -317,3 +337,128 @@ def timedeltas(
     if min_value == max_value:
         return just(min_value)
     return TimedeltaStrategy(min_value=min_value, max_value=max_value)
+
+
+@lru_cache(maxsize=None)
+def _valid_key_cacheable(tzpath, key):
+    assert isinstance(tzpath, tuple)  # zoneinfo changed, better update this function!
+    for root in tzpath:
+        if os.path.exists(os.path.join(root, key)):  # pragma: no branch
+            # No branch because most systems only have one TZPATH component.
+            return True
+    else:  # pragma: no cover
+        # This branch is only taken for names which are known to zoneinfo
+        # but not present on the filesystem, i.e. on Windows with tzdata,
+        # and so is never executed by our coverage tests.
+        if importlib_resources is None:
+            raise ImportError(
+                "The importlib_resources module is required, but could not be "
+                "imported.  Run `pip install hypothesis[zoneinfo]` and try again."
+            )
+        *package_loc, resource_name = key.split("/")
+        package = "tzdata.zoneinfo." + ".".join(package_loc)
+        try:
+            return importlib_resources.is_resource(package, resource_name)
+        except ModuleNotFoundError:
+            return False
+
+
+@defines_strategy(force_reusable_values=True)
+def timezone_keys(
+    *,
+    # allow_alias: bool = True,
+    # allow_deprecated: bool = True,
+    allow_prefix: bool = True,
+) -> SearchStrategy[str]:
+    """A strategy for :wikipedia:`IANA timezone names <List_of_tz_database_time_zones>`.
+
+    As well as timezone names like ``"UTC"``, ``"Australia/Sydney"``, or
+    ``"America/New_York"``, this strategy can generate:
+
+    - Aliases such as ``"Antarctica/McMurdo"``, which links to ``"Pacific/Auckland"``.
+    - Deprecated names such as ``"Antarctica/South_Pole"``, which *also* links to
+      ``"Pacific/Auckland"``.  Note that most but
+      not all deprecated timezone names are also aliases.
+    - Timezone names with the ``"posix/"`` or ``"right/"`` prefixes, unless
+      ``allow_prefix=False``.
+
+    These strings are provided separately from Tzinfo objects - such as ZoneInfo
+    instances from the timezones() strategy - to facilitate testing of timezone
+    logic without needing workarounds to access non-canonical names.
+
+    .. note::
+
+        The :mod:`python:zoneinfo` module is new in Python 3.9, so you will need
+        to install the :pypi:`backports.zoneinfo` module on earlier versions, and
+        the :pypi:`importlib_resources` backport on Python 3.6.
+
+        ``pip install hypothesis[zoneinfo]`` will install these conditional
+        dependencies if and only if they are needed.
+
+    On Windows, you may need to access IANA timezone data via the :pypi:`tzdata`
+    package.  For non-IANA timezones, such as Windows-native names or GNU TZ
+    strings, we recommend using :func:`~hypothesis.strategies.sampled_from` with
+    the :pypi:`dateutil <python-dateutil>` package, e.g.
+    :meth:`dateutil:dateutil.tz.tzwin.list`.
+    """
+    # check_type(bool, allow_alias, "allow_alias")
+    # check_type(bool, allow_deprecated, "allow_deprecated")
+    check_type(bool, allow_prefix, "allow_prefix")
+    if zoneinfo is None:  # pragma: no cover
+        raise ModuleNotFoundError(
+            "The zoneinfo module is required, but could not be imported.  "
+            "Run `pip install hypothesis[zoneinfo]` and try again."
+        )
+
+    available_timezones = ("UTC",) + tuple(sorted(zoneinfo.available_timezones()))
+
+    # TODO: filter out alias and deprecated names if disallowed
+
+    # When prefixes are allowed, we first choose a key and then flatmap to get our
+    # choice with one of the available prefixes.  That in turn means that we need
+    # some logic to determine which prefixes are available for a given key:
+
+    def valid_key(key):
+        return key == "UTC" or _valid_key_cacheable(zoneinfo.TZPATH, key)
+
+    # TODO: work out how to place a higher priority on "weird" timezones
+    # For details see https://github.com/HypothesisWorks/hypothesis/issues/2414
+    strategy = sampled_from([key for key in available_timezones if valid_key(key)])
+
+    if not allow_prefix:
+        return strategy
+
+    def sample_with_prefixes(zone):
+        keys_with_prefixes = (zone, f"posix/{zone}", f"right/{zone}")
+        return sampled_from([key for key in keys_with_prefixes if valid_key(key)])
+
+    return strategy.flatmap(sample_with_prefixes)
+
+
+@defines_strategy(force_reusable_values=True)
+def timezones(*, no_cache: bool = False) -> SearchStrategy["zoneinfo.ZoneInfo"]:
+    """A strategy for :class:`python:zoneinfo.ZoneInfo` objects.
+
+    If ``no_cache=True``, the generated instances are constructed using
+    :meth:`ZoneInfo.no_cache <python:zoneinfo.ZoneInfo.no_cache>` instead
+    of the usual constructor.  This may change the semantics of your datetimes
+    in surprising ways, so only use it if you know that you need to!
+
+    .. note::
+
+        The :mod:`python:zoneinfo` module is new in Python 3.9, so you will need
+        to install the :pypi:`backports.zoneinfo` module on earlier versions, and
+        the :pypi:`importlib_resources` backport on Python 3.6.
+
+        ``pip install hypothesis[zoneinfo]`` will install these conditional
+        dependencies if and only if they are needed.
+    """
+    check_type(bool, no_cache, "no_cache")
+    if zoneinfo is None:  # pragma: no cover
+        raise ModuleNotFoundError(
+            "The zoneinfo module is required, but could not be imported.  "
+            "Run `pip install hypothesis[zoneinfo]` and try again."
+        )
+    return timezone_keys().map(
+        zoneinfo.ZoneInfo.no_cache if no_cache else zoneinfo.ZoneInfo
+    )
